/*
 * Copyright(c) 1995,1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: child.c,v 2.3 1997/12/15 16:13:02 gena Exp $
 */

#include <fmail.h>
#include <umail.h>

#ifdef	HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef	HAVE_VFORK_H
#include <vfork.h>
#endif

static struct _proc_info *procs[MAX_CHILDREN];
static int pprogress = 0, runprocs = 0, finishprocs = 0;

#define	SEL_SEC	10

void
close_procs()
{
int i;
char xbuf[64];

 for (i = 0; i < MAX_CHILDREN; i++) {
  if ((procs[i] == NULL) ||
	(procs[i]->pid <= 0))
	continue;

  if (kill(procs[i]->pid, SIGTERM) == -1)
	display_msg(MSG_WARN, "close procs", "Failed to kill pid %d", procs[i]->pid);

  if (procs[i] == NULL)
	continue;

  sprintf(xbuf, "%s/xfexec.%d", tmpdir, procs[i]->pid);
  unlink(xbuf);

  if (procs[i]->ifd > 0)
	close(procs[i]->ifd);

  if (procs[i]->ofd > 0)
	close(procs[i]->ofd);

  if (procs[i]->efd > 0)
	close(procs[i]->efd);

   if (procs[i]->fd_in[1] > 0)
	close(procs[i]->fd_in[1]);

   if (procs[i]->fd_out[0] > 0)
	close(procs[i]->fd_out[0]);

   if (procs[i]->fd_err[0] > 0)
	close(procs[i]->fd_err[0]);

  if (procs[i]->handle)
	(procs[i]->handle)(procs[i]);

  free(procs[i]);
  procs[i] = NULL;
				    }

 runprocs = 0;
 finishprocs = 0;
 pprogress = 0;
}

void
init_procs()
{
int i;

 for (i = 0; i < MAX_CHILDREN; i++)
   procs[i] = NULL;

 runprocs = 0;
 finishprocs = 0;
 pprogress = 0;
}

void
init_pinfo(pinfo)
struct _proc_info *pinfo;
{
 if (!pinfo)
	return;

 pinfo->pid = -1;
 pinfo->status = 0;
 pinfo->wait = WAIT_ASYNC;
 pinfo->u_data = NULL;
 pinfo->ul_data = 0;
 pinfo->handle = NULL;
 pinfo->init = NULL;
 pinfo->ifd = -1;
 pinfo->ofd = -1;
 pinfo->efd = -1;
 pinfo->fd_in[0] = -1;
 pinfo->fd_in[1] = -1;
 pinfo->fd_out[0] = -1;
 pinfo->fd_out[1] = -1;
 pinfo->fd_err[0] = -1;
 pinfo->fd_err[1] = -1;

 return;
}

struct _proc_info *
dup_pid_info(pinfo)
struct _proc_info *pinfo;
{
struct _proc_info *npinfo;

  if ((npinfo = (struct _proc_info *)malloc(sizeof(struct _proc_info))) == NULL) {
	display_msg(MSG_FATAL, "dup_pid_info", "malloc failed");
	return NULL;
 }	

  if (pinfo) 
	memcpy(npinfo, pinfo, sizeof(struct _proc_info));
  else
	init_pinfo(npinfo);

  npinfo->status = -2;
  npinfo->pid = -1;
  return npinfo;
}

RETSIGTYPE
handler(ret)
int ret;
{    
int i, status, expid;

pprogress = 1;

if ((expid = wait(&status)) == -1) 
	status = -1;

for (i = 0; i < MAX_CHILDREN; i++) {
 if (procs[i] &&
	((procs[i]->pid == expid) || (procs[i]->pid == 0))) {
	procs[i]->status = status;
	procs[i]->wait |= WAIT_FINISHED;
	if (runprocs)
		runprocs--;
	finishprocs++;
	break;			   			    }
				   }

signal(SIGCHLD, handler);
pprogress = 0;

return;
}

void
check_extprocs()
{    
int i, status;
char xbuf[64], pbuf[32], buf[255];
FILE *xfd;

if (pprogress || !finishprocs)
	return;

for (i = 0; i < MAX_CHILDREN; i++) {
 if (procs[i] && (procs[i]->wait & WAIT_FINISHED)) {
   if (finishprocs)
	finishprocs--;
   status = procs[i]->status;
   procs[i]->wait &= ~WAIT_FINISHED;
   sprintf(xbuf, "%s/xfexec.%d", tmpdir, procs[i]->pid);
   sprintf(pbuf, "pid %d", procs[i]->pid);
   if ((xfd = fopen(xbuf, "r")) != NULL) {
    while (fgets(buf, 255, xfd)) {
	strip_newline(buf);
	display_msg(MSG_LOG, pbuf, "%-.127s", buf);
				 }
    fclose(xfd);
					}
   unlink(xbuf);

   if (procs[i]->ifd > 0)
	close(procs[i]->ifd);

   if (procs[i]->ofd > 0)
	close(procs[i]->ofd);

   if (procs[i]->efd > 0)
	close(procs[i]->efd);

/*
   if (procs[i]->fd_in[1] > 0)
	close(procs[i]->fd_in[1]);
*/

   if (status == -1)
	display_msg(MSG_WARN, "exec", "Child process was not terminated normally (See log for output)");
   else
   if (WIFEXITED(status) && (procs[i]->fd_err[0]==-1))  {
	if (WEXITSTATUS(status) != 0)
                display_msg(MSG_WARN, "exec", "Child process terminated with status %d (See log for output)", WEXITSTATUS(status));
			   				}

   procs[i]->status = WEXITSTATUS(status);
   if (procs[i]->handle)
	(procs[i]->handle)(procs[i]);

   if (procs[i]->wait == WAIT_ASYNC)	{
	free(procs[i]);
	procs[i] = NULL;		}
					   }
				   }

return;
}

int
exec_child(str, pinfo)
char *str;
struct _proc_info *pinfo;
{
struct _proc_info *npinfo;
char buf[2048];
char **namelist;
char **ap;
char *p, c;
int params;
int i, pnum;
int nfd;

 if (!str || !*str || pprogress)
	return -1;

 for (pnum = 0; pnum < MAX_CHILDREN; pnum++) {
  if (procs[pnum] == NULL)
	break;
				             }

 if (procs[pnum] != NULL)  {
	display_msg(MSG_WARN, "exec_child", "Can not run more then %d child processes", MAX_CHILDREN);
	return -1;
			   }

 if ((procs[pnum] = dup_pid_info(pinfo)) == NULL)
	return -1;

 npinfo = procs[pnum];

 if (strlen(str) > sizeof(buf)) {
	display_msg(MSG_WARN, "exec", "too many arguments");
	free(procs[pnum]);
	procs[pnum] = NULL;
	return -1;
			        }

 str = rem_tr_space(str);
 p = str;

 params = 1;
 i = 0;
 buf[0] = '\0';

 while(*p != '\0') {
   c = *p;
   switch (c) {
	case ' ':
	 buf[i] = '\0';
	 i++;
	 params++;
	 p++;
	 while (*p == ' ')
		p++;

	break;

	case '\\':
	 p++;
	 if ((*p == '\\') || (*p == '"') || (*p == '\'')) {
		buf[i] = '\\';
		i++;
		p++;
		break;
							  }
	 buf[i] = c;
	 i++;
	break;

	case '\'':
	case '"':
	 buf[i] = c;
	 i++;
	 p++;

	 if (strchr(p, c) == NULL)
	  break;

	 while (*p != c) {
	  buf[i] = *p;
	  i++;
	  p++;
			 }

	 buf[i] = *p;
	 i++;
	 p++;
	break;

	default:
	 buf[i] = c;
	 i++;
	 p++;
	break;
		}
		   }

 if (!*buf) {
	display_msg(MSG_WARN, "exec", "empty command");
	free(procs[pnum]);
	procs[pnum] = NULL;
	return -1;
	    }

 buf[i] = '\0';
 buf[i+1] = '\0';

 params += 2;
 namelist = (char **)malloc(params * sizeof(char *));
 ap = namelist;

 p = buf;
 while(*p != '\0') {
   *ap++ = p;
   p += strlen(p);
   p++;
                   }

 *ap = NULL;

 signal(SIGCHLD, handler);

 if (npinfo->fd_in[0] == 0) {
    if (pipe(npinfo->fd_in) == -1) {
	display_msg(MSG_WARN, "exec", "Can not open pipe");
	free(procs[pnum]);
	procs[pnum] = NULL;
	return -1;		   }
    if (pinfo)
	pinfo->fd_in[1] = npinfo->fd_in[1];
			    }

 if (npinfo->fd_out[0] == 0) {
    if (pipe(npinfo->fd_out) == -1) {
	display_msg(MSG_WARN, "exec", "Can not open pipe");
	free(procs[pnum]);
	procs[pnum] = NULL;
	return -1;	       	    }
    if (pinfo)
	pinfo->fd_out[0] = npinfo->fd_out[0];
			    }

 if (npinfo->fd_err[0] == 0) {
    if (pipe(npinfo->fd_err) == -1) {
	display_msg(MSG_WARN, "exec", "Can not open pipe");
	free(procs[pnum]);
	procs[pnum] = NULL;
	return -1;	       	    }
    if (pinfo)
	pinfo->fd_err[0] = npinfo->fd_err[0];
			    }

    npinfo->pid = 0;
    runprocs++;

        npinfo->pid = vfork();
        if (npinfo->pid == -1) {
                display_msg(MSG_WARN, "exec", "Can not fork");
		free(namelist);
		if (runprocs)
		  runprocs--;
                return -1;     }

        if (npinfo->pid == 0) {
		char xbuf[64];

		if (npinfo->fd_in[0] > 0) {
			dup2(npinfo->fd_in[0],0);
			close(npinfo->fd_in[1]);
					  }

		if (npinfo->fd_out[0] > 0) {
			dup2(npinfo->fd_out[1],1);
			close(npinfo->fd_out[0]);
					   }

		if (npinfo->fd_err[0] > 0) {
			dup2(npinfo->fd_err[1],2);
			close(npinfo->fd_err[0]);
					   }

		sprintf(xbuf, "%s/xfexec.%d", tmpdir, (int)getpid());
		if ((nfd = open(xbuf, O_RDWR|O_CREAT|O_TRUNC, 00600)) != -1)   {
		   if ((npinfo->fd_out[0] <= 0) &&
			(npinfo->ofd <= 0))
			dup2(nfd, 1);
		   if ((npinfo->fd_err[0] <= 0) &&
			(npinfo->efd <= 0))
			dup2(nfd, 2);
									       }

		if ((npinfo->fd_in[0] <= 0) && (npinfo->ifd > 0))
			dup2(npinfo->ifd, 0);

		if ((npinfo->fd_out[0] <= 0) && (npinfo->ofd > 0))
			dup2(npinfo->ofd, 1);

		if ((npinfo->fd_err[0] <= 0) && (npinfo->efd > 0))
			dup2(npinfo->efd, 2);

		if (npinfo->init)
			npinfo->init(npinfo);

                signal(SIGHUP, SIG_IGN);
                signal(SIGINT, SIG_IGN);
                signal(SIGQUIT, SIG_IGN);
#ifndef __EMX__
                signal(SIGTTIN, SIG_IGN);
                signal(SIGTTOU, SIG_IGN);
#endif
                execvp(buf, namelist);
		unlink(xbuf);
                _exit(2);
			}

 if (npinfo->fd_in[0] > 0)
	close(npinfo->fd_in[0]);

 if (npinfo->fd_out[1] > 0)
	close(npinfo->fd_out[1]);

 if (npinfo->fd_err[1] > 0)
	close(npinfo->fd_err[1]);

 free(namelist);

 if (npinfo->wait == WAIT_ASYNC)
	return 0;

 my_check_forms(npinfo);

 i = npinfo->status;

 free(procs[pnum]);
 procs[pnum] = NULL;

 return i;
}

void
my_check_forms(npinfo)
struct _proc_info *npinfo;
{
struct timeval t;
#if defined(__linux__) || defined (__EMX__)
fd_set read_fds;
#else
struct fd_set read_fds;
#endif
int xconn = 0;

 if (!npinfo)
	return;

 if (npinfo->wait == WAIT_BG) {
	check_extprocs();
	fl_check_forms();     }

 xconn = ConnectionNumber(fl_display);
 FD_ZERO(&read_fds);
 FD_SET(xconn, &read_fds);

 while (npinfo->status == -2) {
	t.tv_sec = SEL_SEC;
	t.tv_usec = 0;

#if defined(__linux__) | defined (__EMX__)
	if (select(xconn + 1, &read_fds, (fd_set *) 0, (fd_set *) 0, &t) < 0) {
#else
#ifdef	hpux
	if (select(xconn + 1, (int *)&read_fds, (int *)NULL, (int *)NULL, &t) < 0) {
#else
	if (select(xconn + 1, &read_fds, (struct fd_set *) 0, (struct fd_set *) 0, &t) < 0) {
#endif
#endif
		if (errno == EINTR) {
			if (npinfo->wait == WAIT_BG) {
				check_extprocs();
				fl_check_forms();    }
			if (npinfo->status != -2)
			  break;
			else
			  continue;
				    }
	   }
	   if (npinfo->wait == WAIT_BG) {
		check_extprocs();
		if (abortpressed()) {
		  if (npinfo->pid > 0)
			kill(npinfo->pid, SIGKILL);
		  return;	    }
					}
	     			     }

 check_extprocs();
 return;
}

int
my_check_io_forms(fd, type)
int fd, type;
{  
#if defined(__linux__) || defined(__EMX__)
fd_set read_fds, write_fds, except_fds;
#else
struct fd_set read_fds, write_fds, except_fds;
#endif
int xconn = 0;

 if (fd < 0)
	return -1;

 xconn = ConnectionNumber(fl_display);

fwait:
 FD_ZERO(&read_fds);
 FD_ZERO(&write_fds);

 switch(type) {
  case 0:
	FD_SET(fd, &read_fds);
  break;

  case 1:
	FD_SET(fd, &write_fds);
  break;

  case 2:
	FD_SET(fd, &read_fds);
	FD_SET(fd, &write_fds);
  break;
	      }

 FD_ZERO(&except_fds);
 FD_SET(fd, &except_fds);

 FD_SET(xconn, &read_fds);

#ifdef	hpux
 if (select(fd > xconn ? fd + 1 : xconn + 1, (int *)&read_fds, (int *)&write_fds, (int *)&except_fds, NULL) < 0) 	{
#else
 if (select(fd > xconn ? fd + 1 : xconn + 1, &read_fds, &write_fds, &except_fds, NULL) < 0) 	{
#endif
	if (errno == EINTR)
		return 0;

	return -1;
		}


 if (FD_ISSET(fd, &except_fds))
	return -1;

 if (FD_ISSET(xconn, &read_fds)) {
	if (abortpressed())
		return -2;
				 }

 if (FD_ISSET(fd, &read_fds) || FD_ISSET(fd, &write_fds))
	return 0;

 goto fwait;

 return 0;
}

void
my_check_forms2()
{
struct timeval t;
#if defined(__linux__) || defined (__EMX__)
fd_set read_fds;
#else
struct fd_set read_fds;
#endif
int xconn = 0;

 xconn = ConnectionNumber(fl_display);
 FD_ZERO(&read_fds);
 FD_SET(xconn, &read_fds);

 t.tv_sec = 0;
 t.tv_usec = 0;

#if defined(__linux__) | defined (__EMX__)
 if (select(xconn + 1, &read_fds, (fd_set *) 0, (fd_set *) 0, &t) == 1)
#else
#ifdef	hpux
 if (select(xconn + 1, (int *)&read_fds, (int *)NULL, (int *)NULL, &t) == 1)
#else
 if (select(xconn + 1, &read_fds, (struct fd_set *) 0, (struct fd_set *) 0, &t) == 1)
#endif
#endif
	fl_check_forms();

 return;
}

